@@ -135,6 +135,8 @@ struct _ClutterStagePrivate
 
   ClutterPickMode     pick_buffer_mode;
 
+  GHashTable *devices;
+
   guint relayout_pending       : 1;
   guint redraw_pending         : 1;
   guint is_fullscreen          : 1;
@@ -813,6 +815,9 @@ _clutter_stage_do_update (ClutterStage *stage)
 
   priv = stage->priv;
 
+  if (CLUTTER_ACTOR_IN_DESTRUCTION (stage))
+    return FALSE;
+
   /* NB: We need to ensure we have an up to date layout *before* we
    * check or clear the pending redraws flag since a relayout may
    * queue a redraw.
@@ -1202,10 +1207,12 @@ clutter_stage_finalize (GObject *object)
   g_queue_foreach (priv->event_queue, (GFunc)clutter_event_free, NULL);
   g_queue_free (priv->event_queue);
 
-  g_free (stage->priv->title);
+  g_free (priv->title);
 
   g_array_free (priv->paint_volume_stack, TRUE);
 
+  g_hash_table_destroy (priv->devices);
+
   G_OBJECT_CLASS (clutter_stage_parent_class)->finalize (object);
 }
 
@@ -1624,6 +1631,8 @@ clutter_stage_init (ClutterStage *self)
 
   priv->paint_volume_stack =
     g_array_new (FALSE, FALSE, sizeof (ClutterPaintVolume));
+
+  priv->devices = g_hash_table_new (NULL, NULL);
 }
 
 /**
@@ -3375,3 +3384,35 @@ clutter_stage_get_accept_focus (ClutterStage *stage)
 
   return stage->priv->accept_focus;
 }
+
+void
+_clutter_stage_add_device (ClutterStage       *stage,
+                           ClutterInputDevice *device)
+{
+  ClutterStagePrivate *priv = stage->priv;
+
+  if (g_hash_table_lookup (priv->devices, device) != NULL)
+    return;
+
+  g_hash_table_insert (priv->devices, device, GINT_TO_POINTER (1));
+  _clutter_input_device_set_stage (device, stage);
+}
+
+void
+_clutter_stage_remove_device (ClutterStage       *stage,
+                              ClutterInputDevice *device)
+{
+  ClutterStagePrivate *priv = stage->priv;
+
+  _clutter_input_device_set_stage (device, NULL);
+  g_hash_table_remove (priv->devices, device);
+}
+
+gboolean
+_clutter_stage_has_device (ClutterStage       *stage,
+                           ClutterInputDevice *device)
+{
+  ClutterStagePrivate *priv = stage->priv;
+
+  return g_hash_table_lookup (priv->devices, device) != NULL;
+}
