@@ -4,6 +4,7 @@
  * An object oriented GL/GLES Abstraction/Utility Layer
  *
  * Copyright (C) 2007,2008,2009 Intel Corporation.
+ * Copyright (C) 2010 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -37,15 +38,18 @@
 #include "cogl-bitmap-private.h"
 #include "cogl-buffer-private.h"
 #include "cogl-pixel-array-private.h"
+#include "cogl-private.h"
 #include "cogl-texture-private.h"
 #include "cogl-texture-driver.h"
 #include "cogl-texture-2d-sliced-private.h"
 #include "cogl-texture-2d-private.h"
+#include "cogl-texture-rectangle-private.h"
 #include "cogl-sub-texture-private.h"
 #include "cogl-atlas-texture-private.h"
-#include "cogl-material.h"
+#include "cogl-pipeline.h"
 #include "cogl-context.h"
 #include "cogl-handle.h"
+#include "cogl-object-private.h"
 #include "cogl-primitives.h"
 #include "cogl-framebuffer-private.h"
 
@@ -115,6 +119,14 @@ cogl_texture_unref (CoglHandle handle)
   cogl_handle_unref (handle);
 }
 
+void
+_cogl_texture_init (CoglTexture *texture,
+                    const CoglTextureVtable *vtable)
+{
+  texture->vtable = vtable;
+  texture->framebuffers = NULL;
+}
+
 void
 _cogl_texture_free (CoglTexture *texture)
 {
@@ -152,21 +164,19 @@ _cogl_texture_determine_internal_format (CoglPixelFormat src_format,
     return dst_format;
 }
 
-gboolean
+CoglBitmap *
 _cogl_texture_prepare_for_upload (CoglBitmap      *src_bmp,
                                   CoglPixelFormat  dst_format,
                                   CoglPixelFormat *dst_format_out,
-                                  CoglBitmap      *dst_bmp,
-                                  gboolean        *copied_bitmap,
                                   GLenum          *out_glintformat,
                                   GLenum          *out_glformat,
                                   GLenum          *out_gltype)
 {
-  dst_format = _cogl_texture_determine_internal_format (src_bmp->format,
-                                                        dst_format);
+  CoglPixelFormat src_format = _cogl_bitmap_get_format (src_bmp);
+  CoglBitmap *dst_bmp;
 
-  *copied_bitmap = FALSE;
-  *dst_bmp = *src_bmp;
+  dst_format = _cogl_texture_determine_internal_format (src_format,
+                                                        dst_format);
 
   /* OpenGL supports specifying a different format for the internal
      format when uploading texture data. We should use this to convert
@@ -180,26 +190,26 @@ _cogl_texture_prepare_for_upload (CoglBitmap      *src_bmp,
 
   /* If the source format does not have the same premult flag as the
      dst format then we need to copy and convert it */
-  if (_cogl_texture_needs_premult_conversion (src_bmp->format,
+  if (_cogl_texture_needs_premult_conversion (src_format,
                                               dst_format))
     {
-      dst_bmp->data = g_memdup (dst_bmp->data,
-                                dst_bmp->height * dst_bmp->rowstride);
-      *copied_bitmap = TRUE;
+      dst_bmp = _cogl_bitmap_copy (src_bmp);
 
       if (!_cogl_bitmap_convert_premult_status (dst_bmp,
-                                                src_bmp->format ^
+                                                src_format ^
                                                 COGL_PREMULT_BIT))
         {
-          g_free (dst_bmp->data);
-          return FALSE;
+          cogl_object_unref (dst_bmp);
+          return NULL;
         }
     }
+  else
+    dst_bmp = cogl_object_ref (src_bmp);
 
   /* Use the source format from the src bitmap type and the internal
      format from the dst format type so that GL can do the
      conversion */
-  _cogl_pixel_format_to_gl (src_bmp->format,
+  _cogl_pixel_format_to_gl (src_format,
                             NULL, /* internal format */
                             out_glformat,
                             out_gltype);
@@ -212,28 +222,23 @@ _cogl_texture_prepare_for_upload (CoglBitmap      *src_bmp,
   {
     CoglPixelFormat closest_format;
 
-    closest_format = _cogl_pixel_format_to_gl (dst_bmp->format,
+    closest_format = _cogl_pixel_format_to_gl (dst_format,
                                                out_glintformat,
                                                out_glformat,
                                                out_gltype);
 
-
-    if (closest_format != src_bmp->format)
-      {
-        if (!_cogl_bitmap_convert_format_and_premult (src_bmp,
-                                                      dst_bmp,
-                                                      closest_format))
-          return FALSE;
-
-        *copied_bitmap = TRUE;
-      }
+    if (closest_format != src_format)
+      dst_bmp = _cogl_bitmap_convert_format_and_premult (src_bmp,
+                                                         closest_format);
+    else
+      dst_bmp = cogl_object_ref (src_bmp);
   }
 #endif /* HAVE_COGL_GL */
 
   if (dst_format_out)
     *dst_format_out = dst_format;
 
-  return TRUE;
+  return dst_bmp;
 }
 
 void
@@ -262,19 +267,19 @@ _cogl_texture_prep_gl_alignment_for_pixels_download (int pixels_rowstride)
     GE( glPixelStorei (GL_PACK_ALIGNMENT, 1) );
 }
 
-/* FIXME: wrap modes should be set on materials not textures */
+/* FIXME: wrap modes should be set on pipelines not textures */
 void
 _cogl_texture_set_wrap_mode_parameters (CoglHandle handle,
                                         GLenum wrap_mode_s,
                                         GLenum wrap_mode_t,
-                                        GLenum wrap_mode_r)
+                                        GLenum wrap_mode_p)
 {
   CoglTexture *tex = COGL_TEXTURE (handle);
 
   tex->vtable->set_wrap_mode_parameters (tex,
                                          wrap_mode_s,
                                          wrap_mode_t,
-                                         wrap_mode_r);
+                                         wrap_mode_p);
 }
 
 /* This is like CoglSpanIter except it deals with floats and it
@@ -402,7 +407,8 @@ cogl_texture_new_from_data (unsigned int      width,
 			    unsigned int      rowstride,
 			    const guint8     *data)
 {
-  CoglBitmap bitmap;
+  CoglBitmap *bmp;
+  CoglHandle tex;
 
   if (format == COGL_PIXEL_FORMAT_ANY)
     return COGL_INVALID_HANDLE;
@@ -415,13 +421,18 @@ cogl_texture_new_from_data (unsigned int      width,
     rowstride = width * _cogl_get_format_bpp (format);
 
   /* Wrap the data into a bitmap */
-  bitmap.width = width;
-  bitmap.height = height;
-  bitmap.data = (guint8 *) data;
-  bitmap.format = format;
-  bitmap.rowstride = rowstride;
+  bmp = _cogl_bitmap_new_from_data ((guint8 *) data,
+                                    format,
+                                    width,
+                                    height,
+                                    rowstride,
+                                    NULL, NULL);
 
-  return cogl_texture_new_from_bitmap (&bitmap, flags, internal_format);
+  tex = cogl_texture_new_from_bitmap (bmp, flags, internal_format);
+
+  cogl_object_unref (bmp);
+
+  return tex;
 }
 
 CoglHandle
@@ -455,29 +466,29 @@ cogl_texture_new_from_file (const char        *filename,
                             CoglPixelFormat    internal_format,
                             GError           **error)
 {
-  CoglHandle bmp_handle;
   CoglBitmap *bmp;
   CoglHandle handle = COGL_INVALID_HANDLE;
+  CoglPixelFormat src_format;
 
   g_return_val_if_fail (error == NULL || *error == NULL, COGL_INVALID_HANDLE);
 
-  bmp_handle = cogl_bitmap_new_from_file (filename, error);
-  if (bmp_handle == COGL_INVALID_HANDLE)
+  bmp = cogl_bitmap_new_from_file (filename, error);
+  if (bmp == NULL)
     return COGL_INVALID_HANDLE;
 
-  bmp = (CoglBitmap *) bmp_handle;
+  src_format = _cogl_bitmap_get_format (bmp);
 
   /* We know that the bitmap data is solely owned by this function so
      we can do the premult conversion in place. This avoids having to
      copy the bitmap which will otherwise happen in
      _cogl_texture_prepare_for_upload */
-  internal_format = _cogl_texture_determine_internal_format (bmp->format,
-                                                             internal_format);
-  if (!_cogl_texture_needs_premult_conversion (bmp->format, internal_format) ||
-      _cogl_bitmap_convert_premult_status (bmp, bmp->format ^ COGL_PREMULT_BIT))
+  internal_format =
+    _cogl_texture_determine_internal_format (src_format, internal_format);
+  if (!_cogl_texture_needs_premult_conversion (src_format, internal_format) ||
+      _cogl_bitmap_convert_premult_status (bmp, src_format ^ COGL_PREMULT_BIT))
     handle = cogl_texture_new_from_bitmap (bmp, flags, internal_format);
 
-  cogl_handle_unref (bmp);
+  cogl_object_unref (bmp);
 
   return handle;
 }
@@ -491,15 +502,38 @@ cogl_texture_new_from_foreign (GLuint           gl_handle,
 			       GLuint           y_pot_waste,
 			       CoglPixelFormat  format)
 {
-  /* FIXME: only create a sliced texture if x or y waste was specified
-   */
-  return _cogl_texture_2d_sliced_new_from_foreign (gl_handle,
-                                                   gl_target,
-                                                   width,
-                                                   height,
-                                                   x_pot_waste,
-                                                   y_pot_waste,
-                                                   format);
+#if HAVE_COGL_GL
+  if (gl_target == GL_TEXTURE_RECTANGLE_ARB)
+    {
+      if (x_pot_waste != 0 || y_pot_waste != 0)
+        {
+          /* It shouldn't be necessary to have waste in this case since
+           * the texture isn't limited to power of two sizes. */
+          g_warning ("You can't create a foreign GL_TEXTURE_RECTANGLE cogl "
+                     "texture with waste\n");
+          return COGL_INVALID_HANDLE;
+        }
+
+      return _cogl_texture_rectangle_new_from_foreign (gl_handle,
+                                                       width,
+                                                       height,
+                                                       format);
+    }
+#endif
+
+  if (x_pot_waste != 0 || y_pot_waste != 0)
+    return _cogl_texture_2d_sliced_new_from_foreign (gl_handle,
+                                                     gl_target,
+                                                     width,
+                                                     height,
+                                                     x_pot_waste,
+                                                     y_pot_waste,
+                                                     format);
+  else
+    return _cogl_texture_2d_new_from_foreign (gl_handle,
+                                              width,
+                                              height,
+                                              format);
 }
 
 gboolean
@@ -541,6 +575,7 @@ cogl_texture_new_from_buffer_EXP (CoglHandle          buffer,
   CoglHandle texture;
   CoglBuffer *cogl_buffer;
   CoglPixelArray *pixel_array;
+  CoglBitmap *bmp;
 
   g_return_val_if_fail (cogl_is_buffer (buffer), COGL_INVALID_HANDLE);
 
@@ -569,34 +604,16 @@ cogl_texture_new_from_buffer_EXP (CoglHandle          buffer,
       return COGL_INVALID_HANDLE;
     }
 
-#if !defined (COGL_HAS_GLES)
-  if (cogl_features_available (COGL_FEATURE_PBOS))
-    {
-      CoglBitmap bitmap;
+  /* Wrap the buffer into a bitmap */
+  bmp = _cogl_bitmap_new_from_buffer (cogl_buffer,
+                                      format,
+                                      width, height,
+                                      rowstride,
+                                      offset);
 
-      /* Wrap the data into a bitmap */
-      bitmap.width = width;
-      bitmap.height = height;
-      bitmap.data = GUINT_TO_POINTER (offset);
-      bitmap.format = format;
-      bitmap.rowstride = rowstride;
+  texture = cogl_texture_new_from_bitmap (bmp, flags, internal_format);
 
-      _cogl_buffer_bind (cogl_buffer,
-                         COGL_BUFFER_BIND_TARGET_PIXEL_UNPACK);
-      texture =  cogl_texture_new_from_bitmap (&bitmap, flags, internal_format);
-      _cogl_buffer_unbind (cogl_buffer);
-    }
-  else
-#endif
-    {
-      texture = cogl_texture_new_from_data (width,
-			                    height,
-                                            flags,
-                                            format,
-                                            internal_format,
-                                            rowstride,
-                                            cogl_buffer->data);
-    }
+  cogl_object_unref (bmp);
 
   return texture;
 }
@@ -823,8 +840,6 @@ _cogl_texture_set_region_from_bitmap (CoglHandle    handle,
                                       CoglBitmap   *bmp)
 {
   CoglTexture *tex = COGL_TEXTURE (handle);
-  CoglBitmap   tmp_bmp;
-  gboolean     tmp_bmp_owner = FALSE;
   GLenum       closest_gl_format;
   GLenum       closest_gl_type;
   gboolean     ret;
@@ -835,24 +850,20 @@ _cogl_texture_set_region_from_bitmap (CoglHandle    handle,
 
   /* Prepare the bitmap so that it will do the premultiplication
      conversion */
-  _cogl_texture_prepare_for_upload (bmp,
-                                    cogl_texture_get_format (handle),
-                                    NULL,
-                                    &tmp_bmp,
-                                    &tmp_bmp_owner,
-                                    NULL,
-                                    &closest_gl_format,
-                                    &closest_gl_type);
+  bmp = _cogl_texture_prepare_for_upload (bmp,
+                                          cogl_texture_get_format (handle),
+                                          NULL,
+                                          NULL,
+                                          &closest_gl_format,
+                                          &closest_gl_type);
 
   ret = tex->vtable->set_region (handle,
                                  src_x, src_y,
                                  dst_x, dst_y,
                                  dst_width, dst_height,
-                                 &tmp_bmp);
+                                 bmp);
 
-  /* Free data if owner */
-  if (tmp_bmp_owner)
-    g_free (tmp_bmp.data);
+  cogl_object_unref (bmp);
 
   return ret;
 }
@@ -871,28 +882,35 @@ cogl_texture_set_region (CoglHandle       handle,
 			 unsigned int     rowstride,
 			 const guint8    *data)
 {
-  int              bpp;
-  CoglBitmap       source_bmp;
+  CoglBitmap *source_bmp;
+  gboolean    ret;
 
   /* Check for valid format */
   if (format == COGL_PIXEL_FORMAT_ANY)
     return FALSE;
 
-  /* Init source bitmap */
-  source_bmp.width = width;
-  source_bmp.height = height;
-  source_bmp.format = format;
-  source_bmp.data = (guint8 *) data;
-
   /* Rowstride from width if none specified */
-  bpp = _cogl_get_format_bpp (format);
-  source_bmp.rowstride = (rowstride == 0) ? width * bpp : rowstride;
+  if (rowstride == 0)
+    rowstride = _cogl_get_format_bpp (format) * width;
 
-  return _cogl_texture_set_region_from_bitmap (handle,
-                                               src_x, src_y,
-                                               dst_x, dst_y,
-                                               dst_width, dst_height,
-                                               &source_bmp);
+  /* Init source bitmap */
+  source_bmp = _cogl_bitmap_new_from_data ((guint8 *) data,
+                                           format,
+                                           width,
+                                           height,
+                                           rowstride,
+                                           NULL, /* destroy_fn */
+                                           NULL); /* destroy_fn_data */
+
+  ret = _cogl_texture_set_region_from_bitmap (handle,
+                                              src_x, src_y,
+                                              dst_x, dst_y,
+                                              dst_width, dst_height,
+                                              source_bmp);
+
+  cogl_object_unref (source_bmp);
+
+  return ret;
 }
 
 /* Reads back the contents of a texture by rendering it to the framebuffer
@@ -919,7 +937,7 @@ do_texture_draw_and_read (CoglHandle   handle,
   float       tx1, ty1;
   float       tx2, ty2;
   int         bw,  bh;
-  CoglBitmap  rect_bmp;
+  CoglBitmap  *rect_bmp;
   unsigned int  tex_width, tex_height;
 
   bpp = _cogl_get_format_bpp (COGL_PIXEL_FORMAT_RGBA_8888);
@@ -947,10 +965,19 @@ do_texture_draw_and_read (CoglHandle   handle,
       /* Walk X axis until whole bitmap width consumed */
       for (bw = tex_width; bw > 0; bw-=viewport[2])
         {
+          int width;
+          int height;
+          int rowstride;
+          guint8 *data;
+
           /* Rectangle X coords */
           rx1 = rx2;
           rx2 += (bw < viewport[2]) ? bw : viewport[2];
 
+          width = rx2 - rx1;
+          height = ry2 - ry1;
+          rowstride = width * bpp;
+
           /* Normalized texture X coords */
           tx1 = tx2;
           tx2 = (rx2 / (float) tex_width);
@@ -962,31 +989,35 @@ do_texture_draw_and_read (CoglHandle   handle,
                                               tx1, ty1,
                                               tx2, ty2);
 
+          data = g_malloc (height * rowstride);
+
           /* Read into a temporary bitmap */
-          rect_bmp.format = COGL_PIXEL_FORMAT_RGBA_8888;
-          rect_bmp.width = rx2 - rx1;
-          rect_bmp.height = ry2 - ry1;
-          rect_bmp.rowstride = bpp * rect_bmp.width;
-          rect_bmp.data = g_malloc (rect_bmp.rowstride *
-                                    rect_bmp.height);
+          rect_bmp =
+            _cogl_bitmap_new_from_data (data,
+                                        COGL_PIXEL_FORMAT_RGBA_8888,
+                                        width,
+                                        height,
+                                        rowstride,
+                                        (CoglBitmapDestroyNotify) g_free,
+                                        NULL);
 
           cogl_read_pixels (viewport[0], viewport[1],
-                            rect_bmp.width,
-                            rect_bmp.height,
+                            width,
+                            height,
                             COGL_READ_PIXELS_COLOR_BUFFER,
                             COGL_PIXEL_FORMAT_RGBA_8888_PRE,
-                            rect_bmp.data);
+                            data);
 
           /* Copy to target bitmap */
-          _cogl_bitmap_copy_subregion (&rect_bmp,
+          _cogl_bitmap_copy_subregion (rect_bmp,
                                        target_bmp,
                                        0,0,
                                        rx1,ry1,
-                                       rect_bmp.width,
-                                       rect_bmp.height);
+                                       width,
+                                       height);
 
           /* Free temp bitmap */
-          g_free (rect_bmp.data);
+          cogl_object_unref (rect_bmp);
         }
     }
 }
@@ -1006,10 +1037,12 @@ _cogl_texture_draw_and_read (CoglHandle   handle,
   int        bpp;
   CoglFramebuffer *framebuffer;
   int        viewport[4];
-  CoglBitmap alpha_bmp;
-  CoglHandle prev_source;
+  CoglBitmap *alpha_bmp;
   CoglMatrixStack *projection_stack;
   CoglMatrixStack *modelview_stack;
+  int target_width = _cogl_bitmap_get_width (target_bmp);
+  int target_height = _cogl_bitmap_get_height (target_bmp);
+  int target_rowstride = _cogl_bitmap_get_rowstride (target_bmp);
 
   _COGL_GET_CONTEXT (ctx, FALSE);
 
@@ -1042,27 +1075,26 @@ _cogl_texture_draw_and_read (CoglHandle   handle,
 
   /* Direct copy operation */
 
-  if (ctx->texture_download_material == COGL_INVALID_HANDLE)
+  if (ctx->texture_download_pipeline == COGL_INVALID_HANDLE)
     {
-      ctx->texture_download_material = cogl_material_new ();
-      cogl_material_set_blend (ctx->texture_download_material,
+      ctx->texture_download_pipeline = cogl_pipeline_new ();
+      cogl_pipeline_set_blend (ctx->texture_download_pipeline,
                                "RGBA = ADD (SRC_COLOR, 0)",
                                NULL);
     }
 
-  prev_source = cogl_handle_ref (ctx->source_material);
-  cogl_set_source (ctx->texture_download_material);
+  cogl_push_source (ctx->texture_download_pipeline);
 
-  cogl_material_set_layer (ctx->texture_download_material, 0, handle);
+  cogl_pipeline_set_layer_texture (ctx->texture_download_pipeline, 0, handle);
 
-  cogl_material_set_layer_combine (ctx->texture_download_material,
+  cogl_pipeline_set_layer_combine (ctx->texture_download_pipeline,
                                    0, /* layer */
                                    "RGBA = REPLACE (TEXTURE)",
                                    NULL);
 
-  cogl_material_set_layer_filters (ctx->texture_download_material, 0,
-                                   COGL_MATERIAL_FILTER_NEAREST,
-                                   COGL_MATERIAL_FILTER_NEAREST);
+  cogl_pipeline_set_layer_filters (ctx->texture_download_pipeline, 0,
+                                   COGL_PIPELINE_FILTER_NEAREST,
+                                   COGL_PIPELINE_FILTER_NEAREST);
 
   do_texture_draw_and_read (handle, target_bmp, viewport);
 
@@ -1086,53 +1118,223 @@ _cogl_texture_draw_and_read (CoglHandle   handle,
       guint8 *srcpixel;
       guint8 *dstpixel;
       int     x,y;
+      int     alpha_rowstride = bpp * target_width;
+
+      if ((dstdata = _cogl_bitmap_map (target_bmp,
+                                       COGL_BUFFER_ACCESS_WRITE,
+                                       COGL_BUFFER_MAP_HINT_DISCARD)) == NULL)
+        return FALSE;
+
+      srcdata = g_malloc (alpha_rowstride * target_height);
 
       /* Create temp bitmap for alpha values */
-      alpha_bmp.format = COGL_PIXEL_FORMAT_RGBA_8888;
-      alpha_bmp.width = target_bmp->width;
-      alpha_bmp.height = target_bmp->height;
-      alpha_bmp.rowstride = bpp * alpha_bmp.width;
-      alpha_bmp.data = g_malloc (alpha_bmp.rowstride *
-                                 alpha_bmp.height);
+      alpha_bmp = _cogl_bitmap_new_from_data (srcdata,
+                                              COGL_PIXEL_FORMAT_RGBA_8888,
+                                              target_width, target_height,
+                                              alpha_rowstride,
+                                              (CoglBitmapDestroyNotify) g_free,
+                                              NULL);
 
       /* Draw alpha values into RGB channels */
-      cogl_material_set_layer_combine (ctx->texture_download_material,
+      cogl_pipeline_set_layer_combine (ctx->texture_download_pipeline,
                                        0, /* layer */
                                        "RGBA = REPLACE (TEXTURE[A])",
                                        NULL);
 
-      do_texture_draw_and_read (handle, &alpha_bmp, viewport);
+      do_texture_draw_and_read (handle, alpha_bmp, viewport);
 
       /* Copy temp R to target A */
-      srcdata = alpha_bmp.data;
-      dstdata = target_bmp->data;
 
-      for (y=0; y<target_bmp->height; ++y)
+      for (y=0; y<target_height; ++y)
         {
-          for (x=0; x<target_bmp->width; ++x)
+          for (x=0; x<target_width; ++x)
             {
               srcpixel = srcdata + x*bpp;
               dstpixel = dstdata + x*bpp;
               dstpixel[3] = srcpixel[0];
             }
-          srcdata += alpha_bmp.rowstride;
-          dstdata += target_bmp->rowstride;
+          srcdata += alpha_rowstride;
+          dstdata += target_rowstride;
         }
 
-      g_free (alpha_bmp.data);
+      _cogl_bitmap_unmap (target_bmp);
+
+      cogl_object_unref (alpha_bmp);
     }
 
   /* Restore old state */
   _cogl_matrix_stack_pop (modelview_stack);
   _cogl_matrix_stack_pop (projection_stack);
 
-  /* restore the original material */
-  cogl_set_source (prev_source);
-  cogl_handle_unref (prev_source);
+  /* restore the original pipeline */
+  cogl_pop_source ();
 
   return TRUE;
 }
 
+static gboolean
+get_texture_bits_via_offscreen (CoglHandle      texture_handle,
+                                int             x,
+                                int             y,
+                                int             width,
+                                int             height,
+                                guint8         *dst_bits,
+                                unsigned int    dst_rowstride,
+                                CoglPixelFormat dst_format)
+{
+  CoglFramebuffer *framebuffer;
+
+  _COGL_GET_CONTEXT (ctx, FALSE);
+
+  if (!cogl_features_available (COGL_FEATURE_OFFSCREEN))
+    return FALSE;
+
+  framebuffer = _cogl_offscreen_new_to_texture_full
+                                      (texture_handle,
+                                       COGL_OFFSCREEN_DISABLE_DEPTH_AND_STENCIL,
+                                       0);
+
+  if (framebuffer == NULL)
+    return FALSE;
+
+  cogl_push_framebuffer (framebuffer);
+
+  _cogl_read_pixels_with_rowstride (x, y, width, height,
+                                    COGL_READ_PIXELS_COLOR_BUFFER,
+                                    dst_format, dst_bits, dst_rowstride);
+
+  cogl_pop_framebuffer ();
+
+  cogl_object_unref (framebuffer);
+
+  return TRUE;
+}
+
+static gboolean
+get_texture_bits_via_copy (CoglHandle      texture_handle,
+                           int             x,
+                           int             y,
+                           int             width,
+                           int             height,
+                           guint8         *dst_bits,
+                           unsigned int    dst_rowstride,
+                           CoglPixelFormat dst_format)
+{
+  CoglTexture *tex = COGL_TEXTURE (texture_handle);
+  unsigned int full_rowstride;
+  guint8 *full_bits;
+  gboolean ret = TRUE;
+  int bpp;
+  int full_tex_width, full_tex_height;
+
+  full_tex_width = cogl_texture_get_width (texture_handle);
+  full_tex_height = cogl_texture_get_height (texture_handle);
+
+  bpp = _cogl_get_format_bpp (dst_format);
+
+  full_rowstride = bpp * full_tex_width;
+  full_bits = g_malloc (full_rowstride * full_tex_height);
+
+  if (tex->vtable->get_data (tex,
+                             dst_format,
+                             full_rowstride,
+                             full_bits))
+    {
+      guint8 *dst = dst_bits;
+      guint8 *src = full_bits + x * bpp + y * full_rowstride;
+      int i;
+
+      for (i = 0; i < height; i++)
+        {
+          memcpy (dst, src, bpp * width);
+          dst += dst_rowstride;
+          src += full_rowstride;
+        }
+    }
+  else
+    ret = FALSE;
+
+  g_free (full_bits);
+
+  return ret;
+}
+
+typedef struct
+{
+  int         orig_width;
+  int         orig_height;
+  CoglBitmap *target_bmp;
+  guint8     *target_bits;
+  gboolean    success;
+} CoglTextureGetData;
+
+static void
+texture_get_cb (CoglHandle   texture_handle,
+                const float *subtexture_coords,
+                const float *virtual_coords,
+                void        *user_data)
+{
+  CoglTexture *tex = COGL_TEXTURE (texture_handle);
+  CoglTextureGetData *tg_data = user_data;
+  CoglPixelFormat format = _cogl_bitmap_get_format (tg_data->target_bmp);
+  int bpp = _cogl_get_format_bpp (format);
+  unsigned int rowstride = _cogl_bitmap_get_rowstride (tg_data->target_bmp);
+  int subtexture_width = cogl_texture_get_width (texture_handle);
+  int subtexture_height = cogl_texture_get_height (texture_handle);
+
+  int x_in_subtexture = (int) (0.5 + subtexture_width * subtexture_coords[0]);
+  int y_in_subtexture = (int) (0.5 + subtexture_height * subtexture_coords[1]);
+  int width = ((int) (0.5 + subtexture_width * subtexture_coords[2])
+               - x_in_subtexture);
+  int height = ((int) (0.5 + subtexture_height * subtexture_coords[3])
+                - y_in_subtexture);
+  int x_in_bitmap = (int) (0.5 + tg_data->orig_width * virtual_coords[0]);
+  int y_in_bitmap = (int) (0.5 + tg_data->orig_height * virtual_coords[1]);
+
+  guint8 *dst_bits;
+
+  if (!tg_data->success)
+    return;
+
+  dst_bits = tg_data->target_bits + x_in_bitmap * bpp + y_in_bitmap * rowstride;
+
+  /* If we can read everything as a single slice, then go ahead and do that
+   * to avoid allocating an FBO. We'll leave it up to the GL implementation to
+   * do glGetTexImage as efficiently as possible. (GLES doesn't have that,
+   * so we'll fall through) */
+  if (x_in_subtexture == 0 && y_in_subtexture == 0 &&
+      width == subtexture_width && height == subtexture_height)
+    {
+      if (tex->vtable->get_data (tex,
+                                 format,
+                                 rowstride,
+                                 dst_bits))
+        return;
+    }
+
+  /* Next best option is a FBO and glReadPixels */
+  if (get_texture_bits_via_offscreen (texture_handle,
+                                      x_in_subtexture, y_in_subtexture,
+                                      width, height,
+                                      dst_bits,
+                                      rowstride,
+                                      format))
+    return;
+
+  /* Getting ugly: read the entire texture, copy out the part we want */
+  if (get_texture_bits_via_copy (texture_handle,
+                                 x_in_subtexture, y_in_subtexture,
+                                 width, height,
+                                 dst_bits,
+                                 rowstride,
+                                 format))
+    return;
+
+  /* No luck, the caller will fall back to the draw-to-backbuffer and
+   * read implementation */
+  tg_data->success = FALSE;
+}
+
 int
 cogl_texture_get_data (CoglHandle       handle,
 		       CoglPixelFormat  format,
@@ -1146,17 +1348,18 @@ cogl_texture_get_data (CoglHandle       handle,
   int              closest_bpp;
   GLenum           closest_gl_format;
   GLenum           closest_gl_type;
-  CoglBitmap       target_bmp;
-  CoglBitmap       new_bmp;
-  gboolean         success;
+  CoglBitmap      *target_bmp;
+  CoglBitmap      *new_bmp;
   guint8          *src;
   guint8          *dst;
   int              y;
   int              tex_width;
   int              tex_height;
 
+  CoglTextureGetData tg_data;
+
   if (!cogl_is_texture (handle))
-    return FALSE;
+    return 0;
 
   tex = COGL_TEXTURE (handle);
 
@@ -1183,61 +1386,148 @@ cogl_texture_get_data (CoglHandle       handle,
                                                        &closest_gl_type);
   closest_bpp = _cogl_get_format_bpp (closest_format);
 
-  target_bmp.width = tex_width;
-  target_bmp.height = tex_height;
-
   /* Is the requested format supported? */
   if (closest_format == format)
-    {
-      /* Target user data directly */
-      target_bmp.format = format;
-      target_bmp.rowstride = rowstride;
-      target_bmp.data = data;
-    }
+    /* Target user data directly */
+    target_bmp = _cogl_bitmap_new_from_data (data,
+                                             format,
+                                             tex_width,
+                                             tex_height,
+                                             rowstride,
+                                             NULL, NULL);
   else
     {
-      /* Target intermediate buffer */
-      target_bmp.format = closest_format;
-      target_bmp.rowstride = target_bmp.width * closest_bpp;
-      target_bmp.data = g_malloc (target_bmp.height * target_bmp.rowstride);
+      int target_rowstride = tex_width * closest_bpp;
+      guint8 *target_data = g_malloc (tex_height * target_rowstride);
+      target_bmp = _cogl_bitmap_new_from_data (target_data,
+                                               closest_format,
+                                               tex_width,
+                                               tex_height,
+                                               target_rowstride,
+                                               (CoglBitmapDestroyNotify) g_free,
+                                               NULL);
     }
 
-  if (!tex->vtable->get_data (tex,
-                              target_bmp.format,
-                              target_bmp.rowstride,
-                              target_bmp.data))
-    /* XXX: In some cases _cogl_texture_2d_download_from_gl may fail
-     * to read back the texture data; such as for GLES which doesn't
-     * support glGetTexImage, so here we fallback to drawing the
-     * texture and reading the pixels from the framebuffer. */
-    _cogl_texture_draw_and_read (tex, &target_bmp,
+  tg_data.orig_width = tex_width;
+  tg_data.orig_height = tex_height;
+  tg_data.target_bmp = target_bmp;
+  tg_data.target_bits = _cogl_bitmap_map (target_bmp, COGL_BUFFER_ACCESS_WRITE,
+                                          COGL_BUFFER_MAP_HINT_DISCARD);
+  if (tg_data.target_bits == NULL)
+    {
+      cogl_object_unref (target_bmp);
+      return 0;
+    }
+  tg_data.success = TRUE;
+
+  /* Iterating through the subtextures allows piecing together
+   * the data for a sliced texture, and allows us to do the
+   * read-from-framebuffer logic here in a simple fashion rather than
+   * passing offsets down through the code. */
+  _cogl_texture_foreach_sub_texture_in_region (handle,
+                                               0, 0, 1, 1,
+                                               texture_get_cb,
+                                               &tg_data);
+
+  _cogl_bitmap_unmap (target_bmp);
+
+  /* XXX: In some cases _cogl_texture_2d_download_from_gl may fail
+   * to read back the texture data; such as for GLES which doesn't
+   * support glGetTexImage, so here we fallback to drawing the
+   * texture and reading the pixels from the framebuffer. */
+  if (!tg_data.success)
+    _cogl_texture_draw_and_read (tex, target_bmp,
                                  closest_gl_format,
                                  closest_gl_type);
 
   /* Was intermediate used? */
   if (closest_format != format)
     {
+      guint8 *new_bmp_data;
+      int new_bmp_rowstride;
+
       /* Convert to requested format */
-      success = _cogl_bitmap_convert_format_and_premult (&target_bmp,
-                                                         &new_bmp,
+      new_bmp = _cogl_bitmap_convert_format_and_premult (target_bmp,
                                                          format);
 
       /* Free intermediate data and return if failed */
-      g_free (target_bmp.data);
-      if (!success)
+      cogl_object_unref (target_bmp);
+
+      if (new_bmp == NULL)
         return 0;
 
-      /* Copy to user buffer */
-      for (y = 0; y < new_bmp.height; ++y)
+      new_bmp_rowstride = _cogl_bitmap_get_rowstride (new_bmp);
+      new_bmp_data = _cogl_bitmap_map (new_bmp, COGL_BUFFER_ACCESS_WRITE,
+                                       COGL_BUFFER_MAP_HINT_DISCARD);
+
+      if (new_bmp_data == NULL)
         {
-          src = new_bmp.data + y * new_bmp.rowstride;
-          dst = data + y * rowstride;
-          memcpy (dst, src, new_bmp.width * bpp);
+          cogl_object_unref (new_bmp);
+          return 0;
         }
 
+      /* Copy to user buffer */
+      for (y = 0; y < tex_height; ++y)
+        {
+          src = new_bmp_data + y * new_bmp_rowstride;
+          dst = data + y * rowstride;
+          memcpy (dst, src, tex_width * bpp);
+        }
+
+      _cogl_bitmap_unmap (new_bmp);
+
       /* Free converted data */
-      g_free (new_bmp.data);
+      cogl_object_unref (new_bmp);
     }
 
   return byte_size;
 }
+
+static void
+_cogl_texture_framebuffer_destroy_cb (void *user_data,
+                                      void *instance)
+{
+  CoglTexture *tex = user_data;
+  CoglFramebuffer *framebuffer = instance;
+
+  tex->framebuffers = g_list_remove (tex->framebuffers, framebuffer);
+}
+
+void
+_cogl_texture_associate_framebuffer (CoglHandle handle,
+                                     CoglFramebuffer *framebuffer)
+{
+  CoglTexture *tex = COGL_TEXTURE (handle);
+  static CoglUserDataKey framebuffer_destroy_notify_key;
+
+  /* Note: we don't take a reference on the framebuffer here because
+   * that would introduce a circular reference. */
+  tex->framebuffers = g_list_prepend (tex->framebuffers, framebuffer);
+
+  /* Since we haven't taken a reference on the framebuffer we setup
+   * some private data so we will be notified if it is destroyed... */
+  _cogl_object_set_user_data (COGL_OBJECT (framebuffer),
+                              &framebuffer_destroy_notify_key,
+                              tex,
+                              _cogl_texture_framebuffer_destroy_cb);
+}
+
+const GList *
+_cogl_texture_get_associated_framebuffers (CoglHandle handle)
+{
+  CoglTexture *tex = COGL_TEXTURE (handle);
+  return tex->framebuffers;
+}
+
+void
+_cogl_texture_flush_journal_rendering (CoglHandle handle)
+{
+  CoglTexture *tex = COGL_TEXTURE (handle);
+  GList *l;
+
+  /* It could be that a referenced texture is part of a framebuffer
+   * which has an associated journal that must be flushed before it
+   * can be sampled from by the current primitive... */
+  for (l = tex->framebuffers; l; l = l->next)
+    _cogl_framebuffer_flush_journal (l->data);
+}
