@@ -45,7 +45,8 @@
 #include "cogl-bitmap-private.h"
 #include "cogl-texture-private.h"
 #include "cogl-texture-driver.h"
-#include "cogl-vertex-attribute-private.h"
+#include "cogl-attribute-private.h"
+#include "cogl-framebuffer-private.h"
 
 #ifdef HAVE_COGL_GL
 #define glClientActiveTexture ctx->drv.pf_glClientActiveTexture
@@ -145,71 +146,12 @@ cogl_check_extension (const char *name, const char *ext)
   return _cogl_check_extension (name, ext);
 }
 
-/* This version of cogl_clear can be used internally as an alternative
-   to avoid flushing the journal or the framebuffer state. This is
-   needed when doing operations that may be called whiling flushing
-   the journal */
-void
-_cogl_clear (const CoglColor *color, unsigned long buffers)
-{
-  GLbitfield gl_buffers = 0;
-
-  if (buffers & COGL_BUFFER_BIT_COLOR)
-    {
-      GE( glClearColor (cogl_color_get_red_float (color),
-			cogl_color_get_green_float (color),
-			cogl_color_get_blue_float (color),
-			cogl_color_get_alpha_float (color)) );
-      gl_buffers |= GL_COLOR_BUFFER_BIT;
-    }
-
-  if (buffers & COGL_BUFFER_BIT_DEPTH)
-    gl_buffers |= GL_DEPTH_BUFFER_BIT;
-
-  if (buffers & COGL_BUFFER_BIT_STENCIL)
-    gl_buffers |= GL_STENCIL_BUFFER_BIT;
-
-  if (!gl_buffers)
-    {
-      static gboolean shown = FALSE;
-
-      if (!shown)
-        {
-	  g_warning ("You should specify at least one auxiliary buffer "
-                     "when calling cogl_clear");
-        }
-
-      return;
-    }
-
-  GE (glClear (gl_buffers));
-
-  /* This is a debugging variable used to visually display the quad
-     batches from the journal. It is reset here to increase the
-     chances of getting the same colours for each frame during an
-     animation */
-  if (G_UNLIKELY (cogl_debug_flags & COGL_DEBUG_RECTANGLES))
-    {
-      _COGL_GET_CONTEXT (ctxt, NO_RETVAL);
-      ctxt->journal_rectangles_color = 1;
-    }
-}
-
+/* XXX: it's expected that we'll deprecated this with
+ * cogl_framebuffer_clear at some point. */
 void
 cogl_clear (const CoglColor *color, unsigned long buffers)
 {
-  COGL_NOTE (DRAW, "Clear begin");
-
-  _cogl_journal_flush ();
-
-  /* NB: _cogl_framebuffer_flush_state may disrupt various state (such
-   * as the pipeline state) when flushing the clip stack, so should
-   * always be done first when preparing to draw. */
-  _cogl_framebuffer_flush_state (_cogl_get_framebuffer (), 0);
-
-  _cogl_clear (color, buffers);;
-
-  COGL_NOTE (DRAW, "Clear end");
+  _cogl_framebuffer_clear (_cogl_get_framebuffer (), buffers, color);
 }
 
 static gboolean
@@ -340,7 +282,7 @@ cogl_set_backface_culling_enabled (gboolean setting)
     return;
 
   /* Currently the journal can't track changes to backface culling state... */
-  _cogl_journal_flush ();
+  _cogl_framebuffer_flush_journal (_cogl_get_framebuffer ());
 
   ctx->enable_backface_culling = setting;
 }
@@ -537,7 +479,12 @@ cogl_disable_fog (void)
 void
 cogl_flush (void)
 {
-  _cogl_journal_flush ();
+  GList *l;
+
+  _COGL_GET_CONTEXT (ctx, NO_RETVAL);
+
+  for (l = ctx->framebuffers; l; l = l->next)
+    _cogl_framebuffer_flush_journal (l->data);
 }
 
 void
@@ -550,7 +497,7 @@ _cogl_read_pixels_with_rowstride (int x,
                                   guint8 *pixels,
                                   int rowstride)
 {
-  CoglFramebuffer *framebuffer;
+  CoglFramebuffer *framebuffer = _cogl_get_framebuffer ();
   int              framebuffer_height;
   int              bpp;
   CoglBitmap      *bmp;
@@ -563,11 +510,32 @@ _cogl_read_pixels_with_rowstride (int x,
 
   g_return_if_fail (source == COGL_READ_PIXELS_COLOR_BUFFER);
 
-  /* make sure any batched primitives get emitted to the GL driver before
-   * issuing our read pixels... */
-  cogl_flush ();
+  if (width == 1 && height == 1 && !framebuffer->clear_clip_dirty)
+    {
+      /* If everything drawn so far for this frame is still in the
+       * Journal then if all of the rectangles only have a flat
+       * opaque color we have a fast-path for reading a single pixel
+       * that avoids the relatively high cost of flushing primitives
+       * to be drawn on the GPU (considering how simple the geometry
+       * is in this case) and then blocking on the long GPU pipelines
+       * for the result.
+       */
+      if (_cogl_framebuffer_try_fast_read_pixel (framebuffer,
+                                                 x, y, source, format,
+                                                 pixels))
+        return;
+    }
 
-  framebuffer = _cogl_get_framebuffer ();
+  /* make sure any batched primitives get emitted to the GL driver
+   * before issuing our read pixels...
+   *
+   * XXX: Note we currently use cogl_flush to ensure *all* journals
+   * are flushed here and not _cogl_journal_flush because we don't
+   * track the dependencies between framebuffers so we don't know if
+   * the current framebuffer depends on the contents of other
+   * framebuffers which could also have associated journal entries.
+   */
+  cogl_flush ();
 
   _cogl_framebuffer_flush_state (framebuffer, 0);
 
@@ -771,7 +739,7 @@ cogl_begin_gl (void)
   _cogl_flush_face_winding ();
 
   /* Disable any cached vertex arrays */
-  _cogl_vertex_attribute_disable_cached_arrays ();
+  _cogl_attribute_disable_cached_arrays ();
 }
 
 void
@@ -1143,3 +1111,13 @@ _cogl_error_quark (void)
   return g_quark_from_static_string ("cogl-error-quark");
 }
 
+/* Until Cogl becomes responsible for handling swap-buffer requests
+ * this API is used by Clutter to notify us when it issues a
+ * swap-buffer on our behalf. */
+void
+_cogl_swap_buffers_notify (void)
+{
+  _COGL_GET_CONTEXT (ctx, NO_RETVAL);
+
+  _cogl_framebuffer_swap_notify (_cogl_get_framebuffer ());
+}
